@@ -15,12 +15,19 @@
 """Representations specific to the Renesas CC-RX compiler family."""
 
 import os
-import typing
+import typing as T
 
 from ...mesonlib import EnvironmentException
 
-if typing.TYPE_CHECKING:
+if T.TYPE_CHECKING:
     from ...environment import Environment
+    from ...compilers.compilers import Compiler
+else:
+    # This is a bit clever, for mypy we pretend that these mixins descend from
+    # Compiler, so we get all of the methods and attributes defined for us, but
+    # for runtime we make them descend from object (which all classes normally
+    # do). This gives up DRYer type checking, with no runtime impact
+    Compiler = object
 
 ccrx_buildtype_args = {
     'plain': [],
@@ -29,7 +36,7 @@ ccrx_buildtype_args = {
     'release': [],
     'minsize': [],
     'custom': [],
-}  # type: typing.Dict[str, typing.List[str]]
+}  # type: T.Dict[str, T.List[str]]
 
 ccrx_optimization_args = {
     '0': ['-optimize=0'],
@@ -38,59 +45,66 @@ ccrx_optimization_args = {
     '2': ['-optimize=2'],
     '3': ['-optimize=max'],
     's': ['-optimize=2', '-size']
-}  # type: typing.Dict[str, typing.List[str]]
+}  # type: T.Dict[str, T.List[str]]
 
 ccrx_debug_args = {
     False: [],
     True: ['-debug']
-}  # type: typing.Dict[bool, typing.List[str]]
+}  # type: T.Dict[bool, T.List[str]]
 
 
-class CcrxCompiler:
-    def __init__(self):
+class CcrxCompiler(Compiler):
+
+    if T.TYPE_CHECKING:
+        is_cross = True
+        can_compile_suffixes = set()  # type: T.Set[str]
+
+    def __init__(self) -> None:
         if not self.is_cross:
             raise EnvironmentException('ccrx supports only cross-compilation.')
         self.id = 'ccrx'
         # Assembly
-        self.can_compile_suffixes.update('s')
-        default_warn_args = []  # type: typing.List[str]
+        self.can_compile_suffixes.add('src')
+        default_warn_args = []  # type: T.List[str]
         self.warn_args = {'0': [],
                           '1': default_warn_args,
                           '2': default_warn_args + [],
-                          '3': default_warn_args + []}
+                          '3': default_warn_args + []}  # type: T.Dict[str, T.List[str]]
 
-    def get_pic_args(self) -> typing.List[str]:
+    def get_pic_args(self) -> T.List[str]:
         # PIC support is not enabled by default for CCRX,
         # if users want to use it, they need to add the required arguments explicitly
         return []
 
-    def get_buildtype_args(self, buildtype: str) -> typing.List[str]:
+    def get_buildtype_args(self, buildtype: str) -> T.List[str]:
         return ccrx_buildtype_args[buildtype]
 
     def get_pch_suffix(self) -> str:
         return 'pch'
 
-    def get_pch_use_args(self, pch_dir: str, header: str) -> typing.List[str]:
+    def get_pch_use_args(self, pch_dir: str, header: str) -> T.List[str]:
         return []
 
-    # Override CCompiler.get_dependency_gen_args
-    def get_dependency_gen_args(self, outtarget: str, outfile: str) -> typing.List[str]:
+    def thread_flags(self, env: 'Environment') -> T.List[str]:
         return []
 
-    def thread_flags(self, env: 'Environment') -> typing.List[str]:
+    def get_coverage_args(self) -> T.List[str]:
         return []
 
-    def get_coverage_args(self) -> typing.List[str]:
+    def get_no_stdinc_args(self) -> T.List[str]:
         return []
 
-    def get_optimization_args(self, optimization_level: str) -> typing.List[str]:
+    def get_no_stdlib_link_args(self) -> T.List[str]:
+        return []
+
+    def get_optimization_args(self, optimization_level: str) -> T.List[str]:
         return ccrx_optimization_args[optimization_level]
 
-    def get_debug_args(self, is_debug: bool) -> typing.List[str]:
+    def get_debug_args(self, is_debug: bool) -> T.List[str]:
         return ccrx_debug_args[is_debug]
 
     @classmethod
-    def unix_args_to_native(cls, args: typing.List[str]) -> typing.List[str]:
+    def unix_args_to_native(cls, args: T.List[str]) -> T.List[str]:
         result = []
         for i in args:
             if i.startswith('-D'):
@@ -108,7 +122,7 @@ class CcrxCompiler:
             result.append(i)
         return result
 
-    def compute_parameters_with_absolute_paths(self, parameter_list: typing.List[str], build_dir: str) -> typing.List[str]:
+    def compute_parameters_with_absolute_paths(self, parameter_list: T.List[str], build_dir: str) -> T.List[str]:
         for idx, i in enumerate(parameter_list):
             if i[:9] == '-include=':
                 parameter_list[idx] = i[:9] + os.path.normpath(os.path.join(build_dir, i[9:]))
