@@ -452,7 +452,7 @@ class IconvBuiltinDependency(BuiltinDependency):
     def __init__(self, name: str, env: 'Environment', kwargs: T.Dict[str, T.Any]):
         super().__init__(name, env, kwargs)
 
-        if self.clib_compiler.has_function('iconv_open', '', env)[0]:
+        if self.clib_compiler.has_function('iconv_open', '#include <iconv.h>', env)[0]:
             self.is_found = True
 
 
@@ -480,7 +480,7 @@ class IntlSystemDependency(SystemDependency):
         super().__init__(name, env, kwargs)
 
         h = self.clib_compiler.has_header('libintl.h', '', env)
-        self.link_args =  self.clib_compiler.find_library('intl', env, [], self.libtype)
+        self.link_args = self.clib_compiler.find_library('intl', env, [], self.libtype)
 
         if h[0] and self.link_args:
             self.is_found = True
