@@ -18,6 +18,7 @@ import argparse
 import errno
 import os
 import pickle
+import platform
 import shlex
 import shutil
 import subprocess
@@ -251,6 +252,9 @@ def apply_ldconfig(dm: DirMaker) -> None:
         # If we don't have ldconfig, failure is ignored quietly.
         return
 
+    if 'bsd' in platform.system().lower():
+        return
+
     # Try to update ld cache, it could fail if we don't have permission.
     proc, out, err = Popen_safe(['ldconfig', '-v'])
     if proc.returncode == 0:
