@@ -105,13 +105,12 @@ class HotdocTargetBuilder:
                 self.cmd.extend([option, value])
 
     def check_extra_arg_type(self, arg, value):
-        value = getattr(value, 'held_object', value)
         if isinstance(value, list):
             for v in value:
                 self.check_extra_arg_type(arg, v)
             return
 
-        valid_types = (str, bool, mesonlib.File, build.IncludeDirs, build.CustomTarget, build.BuildTarget)
+        valid_types = (str, bool, mesonlib.File, build.IncludeDirs, build.CustomTarget, build.CustomTargetIndex, build.BuildTarget)
         if not isinstance(value, valid_types):
             raise InvalidArguments('Argument "{}={}" should be of type: {}.'.format(
                 arg, value, [t.__name__ for t in valid_types]))
@@ -188,7 +187,6 @@ class HotdocTargetBuilder:
     def process_dependencies(self, deps):
         cflags = set()
         for dep in mesonlib.listify(ensure_list(deps)):
-            dep = getattr(dep, "held_object", dep)
             if isinstance(dep, InternalDependency):
                 inc_args = self.state.get_include_args(dep.include_directories)
                 cflags.update([self.replace_dirs_in_string(x)
@@ -212,6 +210,8 @@ class HotdocTargetBuilder:
                 self.add_extension_paths(dep.extra_extension_paths)
             elif isinstance(dep, build.CustomTarget) or isinstance(dep, build.BuildTarget):
                 self._dependencies.append(dep)
+            elif isinstance(dep, build.CustomTargetIndex):
+                self._dependencies.append(dep.target)
 
         return [f.strip('-I') for f in cflags]
 
@@ -232,7 +232,6 @@ class HotdocTargetBuilder:
     def flatten_config_command(self):
         cmd = []
         for arg in mesonlib.listify(self.cmd, flatten=True):
-            arg = getattr(arg, 'held_object', arg)
             if isinstance(arg, mesonlib.File):
                 arg = arg.absolute_path(self.state.environment.get_source_dir(),
                                         self.state.environment.get_build_dir())
@@ -242,9 +241,12 @@ class HotdocTargetBuilder:
                     cmd.append(os.path.join(self.builddir, arg.get_curdir(), inc_dir))
 
                 continue
-            elif isinstance(arg, build.CustomTarget) or isinstance(arg, build.BuildTarget):
+            elif isinstance(arg, (build.BuildTarget, build.CustomTarget)):
                 self._dependencies.append(arg)
                 arg = self.interpreter.backend.get_target_filename_abs(arg)
+            elif isinstance(arg, build.CustomTargetIndex):
+                self._dependencies.append(arg.target)
+                arg = self.interpreter.backend.get_target_filename_abs(arg)
 
             cmd.append(arg)
 
@@ -265,7 +267,7 @@ class HotdocTargetBuilder:
                 res.append(self.ensure_file(val))
             return res
 
-        if not isinstance(value, mesonlib.File):
+        if isinstance(value, str):
             return mesonlib.File.from_source_file(self.sourcedir, self.subdir, value)
 
         return value
@@ -291,7 +293,7 @@ class HotdocTargetBuilder:
 
     def make_targets(self):
         self.check_forbidden_args()
-        file_types = (str, mesonlib.File)
+        file_types = (str, mesonlib.File, build.CustomTarget, build.CustomTargetIndex)
         self.process_known_arg("--index", file_types, mandatory=True, value_processor=self.ensure_file)
         self.process_known_arg("--project-version", str, mandatory=True)
         self.process_known_arg("--sitemap", file_types, mandatory=True, value_processor=self.ensure_file)
@@ -313,7 +315,7 @@ class HotdocTargetBuilder:
         hotdoc_config_name = fullname + '.json'
         hotdoc_config_path = os.path.join(
             self.builddir, self.subdir, hotdoc_config_name)
-        with open(hotdoc_config_path, 'w') as f:
+        with open(hotdoc_config_path, 'w', encoding='utf-8') as f:
             f.write('{}')
 
         self.cmd += ['--conf-file', hotdoc_config_path]
@@ -357,6 +359,7 @@ class HotdocTargetBuilder:
                 '--builddir', os.path.join(self.builddir, self.subdir)] +
                 self.hotdoc.get_command() +
                 ['run', '--conf-file', hotdoc_config_name])
+            install_script.tag = 'doc'
 
         return (target, install_script)
 
@@ -371,7 +374,7 @@ class HotdocTargetHolder(CustomTargetHolder):
     def config_path_method(self, *args, **kwargs):
         conf = self.held_object.hotdoc_conf.absolute_path(self.interpreter.environment.source_dir,
                                                           self.interpreter.environment.build_dir)
-        return self.interpreter.holderify(conf)
+        return conf
 
 
 class HotdocTarget(build.CustomTarget):
@@ -403,13 +406,15 @@ class HotDocModule(ExtensionModule):
             from hotdoc.run_hotdoc import run  # noqa: F401
             self.hotdoc.run_hotdoc = run
         except Exception as e:
-            raise MesonException('hotdoc {} required but not found. ({})'.format(
-                MIN_HOTDOC_VERSION, e))
+            raise MesonException(f'hotdoc {MIN_HOTDOC_VERSION} required but not found. ({e})')
+        self.methods.update({
+            'has_extensions': self.has_extensions,
+            'generate_doc': self.generate_doc,
+        })
 
     @noKwargs
     def has_extensions(self, state, args, kwargs):
-        res = self.hotdoc.run_hotdoc(['--has-extension=%s' % extension for extension in args]) == 0
-        return ModuleReturnValue(res, [res])
+        return self.hotdoc.run_hotdoc(['--has-extension=%s' % extension for extension in args]) == 0
 
     def generate_doc(self, state, args, kwargs):
         if len(args) != 1:
@@ -419,7 +424,7 @@ class HotDocModule(ExtensionModule):
         project_name = args[0]
         builder = HotdocTargetBuilder(project_name, state, self.hotdoc, self.interpreter, kwargs)
         target, install_script = builder.make_targets()
-        targets = [HotdocTargetHolder(target, self.interpreter)]
+        targets = [target]
         if install_script:
             targets.append(install_script)
 
@@ -427,4 +432,6 @@ class HotDocModule(ExtensionModule):
 
 
 def initialize(interpreter):
-    return HotDocModule(interpreter)
+    mod = HotDocModule(interpreter)
+    mod.interpreter.append_holder_map(HotdocTarget, HotdocTargetHolder)
+    return mod
