@@ -13,10 +13,9 @@
 # limitations under the License.
 
 import os
-import shutil
 import argparse
 import subprocess
-from . import destdir_join
+import typing as T
 
 parser = argparse.ArgumentParser()
 parser.add_argument('command')
@@ -27,26 +26,26 @@ parser.add_argument('--localedir', default='')
 parser.add_argument('--subdir', default='')
 parser.add_argument('--extra-args', default='')
 
-def read_linguas(src_sub):
+def read_linguas(src_sub: str) -> T.List[str]:
     # Syntax of this file is documented here:
     # https://www.gnu.org/software/gettext/manual/html_node/po_002fLINGUAS.html
     linguas = os.path.join(src_sub, 'LINGUAS')
     try:
         langs = []
-        with open(linguas) as f:
+        with open(linguas, encoding='utf-8') as f:
             for line in f:
                 line = line.strip()
                 if line and not line.startswith('#'):
                     langs += line.split()
         return langs
     except (FileNotFoundError, PermissionError):
-        print('Could not find file LINGUAS in {}'.format(src_sub))
+        print(f'Could not find file LINGUAS in {src_sub}')
         return []
 
-def run_potgen(src_sub, pkgname, datadirs, args):
-    listfile = os.path.join(src_sub, 'POTFILES')
+def run_potgen(src_sub: str, pkgname: str, datadirs: str, args: T.List[str]) -> int:
+    listfile = os.path.join(src_sub, 'POTFILES.in')
     if not os.path.exists(listfile):
-        listfile = os.path.join(src_sub, 'POTFILES.in')
+        listfile = os.path.join(src_sub, 'POTFILES')
         if not os.path.exists(listfile):
             print('Could not find file POTFILES in %s' % src_sub)
             return 1
@@ -60,13 +59,7 @@ def run_potgen(src_sub, pkgname, datadirs, args):
                             '-D', os.environ['MESON_SOURCE_ROOT'], '-k_', '-o', ofile] + args,
                            env=child_env)
 
-def gen_gmo(src_sub, bld_sub, langs):
-    for l in langs:
-        subprocess.check_call(['msgfmt', os.path.join(src_sub, l + '.po'),
-                               '-o', os.path.join(bld_sub, l + '.gmo')])
-    return 0
-
-def update_po(src_sub, pkgname, langs):
+def update_po(src_sub: str, pkgname: str, langs: T.List[str]) -> int:
     potfile = os.path.join(src_sub, pkgname + '.pot')
     for l in langs:
         pofile = os.path.join(src_sub, l + '.po')
@@ -76,20 +69,7 @@ def update_po(src_sub, pkgname, langs):
             subprocess.check_call(['msginit', '--input', potfile, '--output-file', pofile, '--locale', l, '--no-translator'])
     return 0
 
-def do_install(src_sub, bld_sub, dest, pkgname, langs):
-    for l in langs:
-        srcfile = os.path.join(bld_sub, l + '.gmo')
-        outfile = os.path.join(dest, l, 'LC_MESSAGES',
-                               pkgname + '.mo')
-        tempfile = outfile + '.tmp'
-        os.makedirs(os.path.dirname(outfile), exist_ok=True)
-        shutil.copyfile(srcfile, tempfile)
-        shutil.copystat(srcfile, tempfile)
-        os.replace(tempfile, outfile)
-        print('Installing %s to %s' % (srcfile, outfile))
-    return 0
-
-def run(args):
+def run(args: T.List[str]) -> int:
     options = parser.parse_args(args)
     subcmd = options.command
     langs = options.langs.split('@@') if options.langs else None
@@ -98,26 +78,16 @@ def run(args):
     if options.subdir:
         subdir = options.subdir
     src_sub = os.path.join(os.environ['MESON_SOURCE_ROOT'], subdir)
-    bld_sub = os.path.join(os.environ['MESON_BUILD_ROOT'], subdir)
 
     if not langs:
         langs = read_linguas(src_sub)
 
     if subcmd == 'pot':
         return run_potgen(src_sub, options.pkgname, options.datadirs, extra_args)
-    elif subcmd == 'gen_gmo':
-        return gen_gmo(src_sub, bld_sub, langs)
     elif subcmd == 'update_po':
         if run_potgen(src_sub, options.pkgname, options.datadirs, extra_args) != 0:
             return 1
         return update_po(src_sub, options.pkgname, langs)
-    elif subcmd == 'install':
-        destdir = os.environ.get('DESTDIR', '')
-        dest = destdir_join(destdir, os.path.join(os.environ['MESON_INSTALL_PREFIX'],
-                                                  options.localedir))
-        if gen_gmo(src_sub, bld_sub, langs) != 0:
-            return 1
-        do_install(src_sub, bld_sub, dest, options.pkgname, langs)
     else:
         print('Unknown subcommand.')
         return 1
